@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2020  Błażej Szczygieł
+    Copyright (C) 2010-2023  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -26,7 +26,6 @@
 #include <GPUInstance.hpp>
 #include <HWDecContext.hpp>
 #include <VideoWriter.hpp>
-#include <QMPlay2OSD.hpp>
 #include <Frame.hpp>
 #include <Settings.hpp>
 #include <Decoder.hpp>
@@ -38,10 +37,16 @@
 #   include <opengl/OpenGLHWInterop.hpp>
 #endif
 #ifdef USE_VULKAN
+#   include <vulkan/VulkanInstance.hpp>
 #   include <vulkan/VulkanHWInterop.hpp>
 #   include <vulkan/VulkanYadifDeint.hpp>
 #endif
 
+extern "C" {
+    #include <libavutil/cpu.h>
+    #include <libavutil/mem.h>
+}
+
 using Functions::gettime;
 using namespace std;
 
@@ -49,6 +54,10 @@ using namespace std;
 #include <QImage>
 #include <QDir>
 
+#ifdef Q_OS_WIN
+#   include <windows.h>
+#endif
+
 #include <cmath>
 
 VideoThr::VideoThr(PlayClass &playC, const QStringList &pluginsName) :
@@ -57,8 +66,7 @@ VideoThr::VideoThr(PlayClass &playC, const QStringList &pluginsName) :
     doScreenshot(false),
     deleteOSD(false), deleteFrame(false), gotFrameOrError(false), decoderError(false),
     W(0), H(0), seq(0),
-    sDec(nullptr),
-    subtitles(nullptr)
+    sDec(nullptr)
 {
     if (QMPlay2Core.renderer() != QMPlay2CoreClass::Renderer::Legacy)
     {
@@ -76,9 +84,10 @@ VideoThr::~VideoThr()
 {
     QMPlay2GUI.videoAdjustment->enableControls();
     QMPlay2GUI.screenSaver->unInhibit(0);
-    delete playC.osd;
-    playC.osd = nullptr;
-    delete subtitles;
+#ifdef Q_OS_WIN
+    setHighTimerResolution<false>();
+#endif
+    playC.osd.reset();
     delete sDec;
 }
 
@@ -102,6 +111,15 @@ bool VideoThr::videoWriterSet()
     return writer->set();
 }
 
+bool VideoThr::lock()
+{
+    if (QThread::currentThread() == thread())
+    {
+        m_subsDisplayLocker = {};
+    }
+    return AVThread::lock();
+}
+
 void VideoThr::stop(bool terminate)
 {
     if (QMPlay2Core.renderer() != QMPlay2CoreClass::Renderer::Legacy)
@@ -128,6 +146,12 @@ void VideoThr::destroySubtitlesDecoder()
         delete sDec;
         sDec = nullptr;
     }
+    m_decodeToAss = false;
+}
+void VideoThr::setSubtitlesDecoder(Decoder *dec, bool decodeToAss)
+{
+    sDec = dec;
+    m_decodeToAss = decodeToAss;
 }
 
 bool VideoThr::setSpherical()
@@ -167,6 +191,14 @@ void VideoThr::setZoom()
     updateSubs();
     writer->modParam("Zoom", playC.zoom);
 }
+
+void VideoThr::setColorInfo(AVColorPrimaries colorPrimaries, AVColorTransferCharacteristic colorTrc)
+{
+    if (writer->hasParam("ColorPrimaries"))
+        writer->modParam("ColorPrimaries", colorPrimaries);
+    if (writer->hasParam("ColorTrc"))
+        writer->modParam("ColorTrc", colorTrc);
+}
 void VideoThr::otherReset()
 {
     if (writer->hasParam("ResetOther"))
@@ -194,6 +226,9 @@ void VideoThr::initFilters()
             if (!QMPlay2Core.isVulkanRenderer())
                 return false;
 
+            if (!static_pointer_cast<QmVk::Instance>(QMPlay2Core.gpuInstance())->checkFiltersSupported())
+                return false;
+
             shared_ptr<VideoFilter> deintFilter = make_shared<QmVk::YadifDeint>(
                 static_pointer_cast<QmVk::HWInterop>(getHWDecContext())
             );
@@ -211,6 +246,26 @@ void VideoThr::initFilters()
         };
 #endif
 
+        auto iterateVideoFilters = [this, &QMPSettings](bool isHw) {
+            for (QString filterName : QMPSettings.getStringList("VideoFilters"))
+            {
+                if (filterName.left(1).toInt()) //if filter is enabled
+                {
+                    bool ok = false;
+                    filterName = filterName.mid(1);
+                    if (auto filter = filters.on(filterName, isHw))
+                    {
+                        filter->modParam("W", W);
+                        filter->modParam("H", H);
+                        if (!(ok = filter->processParams()))
+                            filters.off(filter);
+                    }
+                    if (!ok && W > 0 && H > 0)
+                        QMPlay2Core.logError(tr("Error initializing filter") + " \"" + filterName + '"');
+                }
+            }
+        };
+
         if (getHWDecContext())
         {
             if (auto hwFilter = dec->hwAccelFilter())
@@ -227,6 +282,8 @@ void VideoThr::initFilters()
             {
                 enableVulkanDeint();
             }
+
+            iterateVideoFilters(true);
 #endif
         }
         else
@@ -243,7 +300,7 @@ void VideoThr::initFilters()
             if (enableSoftwareDeint)
             {
                 const QString deintFilterName = QMPSettings.getString("Deinterlace/SoftwareMethod");
-                if (auto deintFilter = filters.on(deintFilterName))
+                if (auto deintFilter = filters.on(deintFilterName, false))
                 {
                     bool ok = false;
                     if (deintFilter->modParam("DeinterlaceFlags", deintFlags))
@@ -261,23 +318,7 @@ void VideoThr::initFilters()
                 }
             }
 
-            for (QString filterName : QMPSettings.getStringList("VideoFilters"))
-            {
-                if (filterName.leftRef(1).toInt()) //if filter is enabled
-                {
-                    bool ok = false;
-                    filterName = filterName.mid(1);
-                    if (auto filter = filters.on(filterName))
-                    {
-                        filter->modParam("W", W);
-                        filter->modParam("H", H);
-                        if (!(ok = filter->processParams()))
-                            filters.off(filter);
-                    }
-                    if (!ok && W > 0 && H > 0)
-                        QMPlay2Core.logError(tr("Error initializing filter") + " \"" + filterName + '"');
-                }
-            }
+            iterateVideoFilters(false);
         }
     }
 
@@ -295,8 +336,10 @@ void VideoThr::updateSubs()
     if (playC.ass)
     {
         playC.subsMutex.lock();
-        if (subtitles)
-            playC.ass->getASS(subtitles, playC.frame_last_pts + playC.frame_last_delay  - playC.subtitlesSync);
+        if (m_subtitlesBusy)
+            playC.ass->getASS(m_subtitlesBusy);
+        if (m_subtitles)
+            playC.ass->getASS(m_subtitles);
         playC.subsMutex.unlock();
     }
 }
@@ -308,7 +351,7 @@ inline VideoWriter *VideoThr::videoWriter() const
 
 void VideoThr::run()
 {
-    bool skip = false, paused = false, oneFrame = false, useLastDelay = false, lastOSDListEmpty = true, maybeFlush = false, lastAVDesync = false, interlaced = false, err = false;
+    bool skip = false, paused = false, oneFrame = false, useLastDelay = false, maybeFlush = false, lastAVDesync = false, interlaced = false, err = false, skipNonKey = false;
     double tmp_time = 0.0, sync_last_pts = 0.0, frame_timer = -1.0, sync_timer = 0.0, framesDisplayedTime = 0.0;
     QMutex emptyBufferMutex;
     Frame videoFrame;
@@ -353,7 +396,9 @@ void VideoThr::run()
 
         if (doScreenshot && !videoFrame.isEmpty())
         {
-            QMetaObject::invokeMethod(this, "screenshot", Q_ARG(Frame, videoFrame));
+            QTimer::singleShot(0, this, [=] {
+                screenshot(videoFrame);
+            });
             doScreenshot = false;
         }
 
@@ -367,7 +412,9 @@ void VideoThr::run()
         {
             if (playC.paused && !paused)
             {
-                QMetaObject::invokeMethod(this, "pause");
+                QTimer::singleShot(0, this, [this] {
+                    pause();
+                });
                 paused = true;
                 frame_timer = -1.0;
                 emit playC.updateBitrateAndFPS(-1, -1, -1.0, 0.0, interlaced); //Set real FPS to 0 on pause
@@ -417,88 +464,25 @@ void VideoThr::run()
             break;
         }
 
-        /* Subtitles */
-        const double subsPts = playC.frame_last_pts + playC.frame_last_delay  - playC.subtitlesSync;
-        QList<const QMPlay2OSD *> osdList, osdListToDelete;
-        playC.subsMutex.lock();
-        const bool canDeleteSubs = (deleteSubs && static_cast<bool>(subtitles));
-        if (sDec) //Image subs (pgssub, dvdsub, ...)
-        {
-            if (!sDec->decodeSubtitle(sPackets, subsPts, subtitles, QSize(W, H), playC.flushVideo))
-            {
-                osdListToDelete += subtitles;
-                subtitles = nullptr;
-            }
-        }
-        else
-        {
-            for (auto &&sPacket : asConst(sPackets))
-            {
-                const QByteArray sPacketData = QByteArray::fromRawData((const char *)sPacket.data(), sPacket.size());
-                if (playC.ass->isASS())
-                    playC.ass->addASSEvent(sPacketData);
-                else
-                    playC.ass->addASSEvent(Functions::convertToASS(sPacketData), sPacket.ts(), sPacket.duration());
-            }
-            if (!playC.ass->getASS(subtitles, subsPts))
-            {
-                osdListToDelete += subtitles;
-                subtitles = nullptr;
-            }
-        }
-        if (subtitles)
-        {
-            const bool hasDuration = subtitles->duration() >= 0.0;
-            if (canDeleteSubs || (subtitles->isStarted() && subsPts < subtitles->pts()) || (hasDuration && subsPts > subtitles->pts() + subtitles->duration()))
-            {
-                osdListToDelete += subtitles;
-                subtitles = nullptr;
-            }
-            else if (subsPts >= subtitles->pts())
-            {
-                subtitles->start();
-                osdList += subtitles;
-            }
-        }
-        playC.subsMutex.unlock();
-        playC.osdMutex.lock();
-        if (playC.osd)
-        {
-            if (deleteOSD || playC.osd->leftDuration() < 0)
-            {
-                osdListToDelete += playC.osd;
-                playC.osd = nullptr;
-            }
-            else
-                osdList += playC.osd;
-        }
-        playC.osdMutex.unlock();
-        if ((!lastOSDListEmpty || !osdList.isEmpty()) && writer->readyWrite())
-        {
-            videoWriter()->writeOSD(osdList);
-            lastOSDListEmpty = osdList.isEmpty();
-        }
-        while (!osdListToDelete.isEmpty())
-            delete osdListToDelete.takeFirst();
-        deleteSubs = deleteOSD = false;
-        /**/
+        const bool flushVideo = playC.flushVideo;
 
         filtersMutex.lock();
-        if (playC.flushVideo)
+        if (flushVideo || skipNonKey)
         {
             filters.clearBuffers();
-            frame_timer = -1.0;
+            if (flushVideo)
+                frame_timer = -1.0;
         }
 
-        if (!packet.isEmpty() || maybeFlush)
+        if ((!packet.isEmpty() || maybeFlush) && (!skipNonKey || packet.hasKeyFrame()))
         {
             Frame decoded;
             AVPixelFormat newPixelFormat = AV_PIX_FMT_NONE;
-            const int bytes_consumed = dec->decodeVideo(packet, decoded, newPixelFormat, playC.flushVideo, skip ? ~0 : (fast >> 1));
+            const int bytes_consumed = dec->decodeVideo(packet, decoded, newPixelFormat, flushVideo || skipNonKey, (skip && !skipNonKey) ? ~0 : (fast >> 1));
             ts = decoded.isTsValid() ? decoded.ts() : qQNaN();
             if (newPixelFormat != AV_PIX_FMT_NONE)
                 emit playC.pixelFormatUpdate(newPixelFormat);
-            if (playC.flushVideo)
+            if (flushVideo)
             {
                 useLastDelay = true; //if seeking
                 playC.flushVideo = false;
@@ -524,7 +508,9 @@ void VideoThr::run()
                 gotFrameOrError = true;
             }
             else if (skip)
+            {
                 filters.removeLastFromInputBuffer();
+            }
             if (bytes_consumed < 0)
             {
                 gotFrameOrError = true;
@@ -534,6 +520,7 @@ void VideoThr::run()
             {
                 tmp_br += bytes_consumed;
             }
+            skipNonKey = false;
         }
 
         // This thread will wait for "DemuxerThr" which'll detect this error and restart with new decoder.
@@ -551,7 +538,83 @@ void VideoThr::run()
             ts = videoFrame.ts();
         filtersMutex.unlock();
 
-        if ((maybeFlush = !qIsNaN(ts)))
+        const bool tsIsNotNan = !qIsNaN(ts);
+
+        /* Subtitles */
+        QMPlay2OSDList osdList;
+        m_subsDisplayMutex.lock(); // Must be locked before "playC.subsMutex"!
+        playC.subsMutex.lock();
+        const double subsPts = tsIsNotNan
+            ? ts - playC.subtitlesSync
+            : qQNaN()
+        ;
+        const bool canDeleteSubs = (deleteSubs && m_subtitles);
+        auto resetSubs = [this] {
+            m_subtitles.reset();
+            m_subtitlesBusy.reset();
+        };
+        if (sDec && !m_decodeToAss) //Image subs (pgssub, dvdsub, ...)
+        {
+            if (!sDec->decodeSubtitle(sPackets, subsPts, m_subtitles, QSize(W, H), flushVideo))
+            {
+                resetSubs();
+            }
+        }
+        else if (playC.ass)
+        {
+            for (auto &&sPacket : qAsConst(sPackets))
+            {
+                if (sDec && m_decodeToAss)
+                {
+                    const auto events = sDec->decodeSubtitle(sPacket);
+                    playC.ass->addASSEvents(events, sPacket.ts(), sPacket.duration());
+                }
+                else
+                {
+                    const auto sPacketData = QByteArray::fromRawData(reinterpret_cast<const char *>(sPacket.data()), sPacket.size());
+                    if (playC.ass->isASS())
+                        playC.ass->addASSEvent(sPacketData);
+                    else
+                        playC.ass->addASSEvent(Functions::convertToASS(sPacketData), sPacket.ts(), sPacket.duration());
+                }
+            }
+            if (tsIsNotNan && !playC.ass->getASS(m_subtitles, subsPts))
+            {
+                resetSubs();
+            }
+        }
+        if (m_subtitles && tsIsNotNan)
+        {
+            const bool hasDuration = m_subtitles->duration() >= 0.0;
+            if (canDeleteSubs || (m_subtitles->isStarted() && subsPts < m_subtitles->pts()) || (hasDuration && subsPts > m_subtitles->pts() + m_subtitles->duration()))
+            {
+                resetSubs();
+            }
+            else if (subsPts >= m_subtitles->pts())
+            {
+                m_subtitles->start();
+                osdList += m_subtitles;
+            }
+        }
+        playC.subsMutex.unlock();
+        m_subsDisplayMutex.unlock();
+        playC.osdMutex.lock();
+        if (playC.osd)
+        {
+            if (deleteOSD || playC.osd->leftDuration() < 0)
+            {
+                playC.osd.reset();
+            }
+            else
+            {
+                osdList += playC.osd;
+            }
+        }
+        playC.osdMutex.unlock();
+        deleteSubs = deleteOSD = false;
+        /**/
+
+        if ((maybeFlush = tsIsNotNan))
         {
             const AVRational currSAR = videoFrame.sampleAspectRatio();
             if (currSAR.num != 0 && currSAR.den != 0 && lastSAR.num != 0 && lastSAR.den != 0 && av_cmp_q(lastSAR, currSAR) != 0) //Aspect ratio has been changed
@@ -582,7 +645,7 @@ void VideoThr::run()
                 playC.chPos(ts);
 
             double delay = ts - playC.frame_last_pts;
-            if (useLastDelay || delay <= 0.0 || (playC.frame_last_pts <= 0.0 && delay > playC.frame_last_delay))
+            if (useLastDelay || delay <= 0.0 || (playC.frame_last_pts <= 0.0 && delay > playC.frame_last_delay) || (delay > 1.0 && delay / playC.frame_last_delay > 10.0))
             {
                 delay = playC.frame_last_delay;
                 useLastDelay = false;
@@ -591,11 +654,11 @@ void VideoThr::run()
             tmp_time += delay;
             ++frames;
 
-            delay /= playC.speed;
-
             playC.frame_last_delay = delay;
             playC.frame_last_pts = ts;
 
+            delay /= playC.speed;
+
             if (playC.skipAudioFrame < 0.0)
                 playC.skipAudioFrame = 0.0;
 
@@ -656,6 +719,8 @@ void VideoThr::run()
                         delay = 0.0;
                         if (fast >= 7)
                             skip = true;
+                        if (fast >= 56 || (fast >= 28 && fDiff >= max_threshold * 4.0))
+                            skipNonKey = true;
                     }
                     else if (diff > 0.0) //obraz idzie za szybko
                     {
@@ -712,8 +777,13 @@ void VideoThr::run()
                 if (!skip && canWrite)
                 {
                     oneFrame = canWrite = false;
-                    QMetaObject::invokeMethod(this, "write", Q_ARG(Frame, videoFrame), Q_ARG(quint32, seq));
-                    if (canSkipFrames)
+                    if (!osdList.isEmpty())
+                        m_subsDisplayLocker = unique_lock<std::mutex>(m_subsDisplayMutex);
+                    QTimer::singleShot(0, this, [=, osdList = move(osdList)]() mutable {
+                        write(videoFrame, move(osdList), seq);
+                        m_subsDisplayLocker = {};
+                    });
+                    if (canSkipFrames && !skipNonKey)
                         ++framesDisplayed;
                 }
                 if (canSkipFrames)
@@ -724,10 +794,39 @@ void VideoThr::run()
         }
 
         mutex.unlock();
+
+        if (br2)
+        {
+            // Take time for another thread to lock the mutex
+            Functions::s_wait(0.001);
+        }
     }
 }
 
-void VideoThr::write(Frame videoFrame, quint32 lastSeq)
+#ifdef Q_OS_WIN
+template<bool h>
+inline void VideoThr::setHighTimerResolution()
+{
+    if (h)
+    {
+        if (!m_timerPrecision)
+        {
+            timeBeginPeriod(1);
+            m_timerPrecision = true;
+        }
+    }
+    else
+    {
+        if (m_timerPrecision)
+        {
+            timeEndPeriod(1);
+            m_timerPrecision = false;
+        }
+    }
+}
+#endif
+
+void VideoThr::write(const Frame &videoFrame, QMPlay2OSDList &&osdList, quint32 lastSeq)
 {
     canWrite = true;
 
@@ -739,7 +838,14 @@ void VideoThr::write(Frame videoFrame, quint32 lastSeq)
         return;
 
     QMPlay2GUI.screenSaver->inhibit(0);
-    videoWriter()->writeVideo(videoFrame);
+#ifdef Q_OS_WIN
+    setHighTimerResolution<true>();
+#endif
+
+    videoWriter()->writeVideo(videoFrame, move(osdList));
+
+    if (m_subsDisplayLocker.owns_lock())
+        swap(m_subtitles, m_subtitlesBusy);
 }
 void VideoThr::screenshot(Frame videoFrame)
 {
@@ -749,10 +855,10 @@ void VideoThr::screenshot(Frame videoFrame)
 #ifdef USE_OPENGL
     if (auto hwGLInterop = dynamic_pointer_cast<OpenGLHWInterop>(getHWDecContext()))
     {
-        img = hwGLInterop->getImage(videoFrame);
+        videoFrame = hwGLInterop->getCpuFrame(videoFrame);
     }
-    else
 #endif
+
     if (imgScaler.create(videoFrame, W, H))
     {
 #ifdef USE_VULKAN
@@ -764,8 +870,11 @@ void VideoThr::screenshot(Frame videoFrame)
                 videoFrame.clear();
         }
 #endif
-        img = QImage(W, H, QImage::Format_RGB32);
-        if (!imgScaler.scale(videoFrame, img.bits()))
+        auto imgData = reinterpret_cast<uint8_t *>(av_malloc(W * H * 4 + av_cpu_max_align()));
+        img = QImage(imgData, W, H, QImage::Format_RGB32, [](void *ptr) {
+            av_free(ptr);
+        }, imgData);
+        if (!imgScaler.scale(videoFrame, imgData))
             img = QImage();
     }
 
@@ -777,17 +886,33 @@ void VideoThr::screenshot(Frame videoFrame)
 
     const QString ext = QMPlay2Core.getSettings().getString("screenshotFormat");
     const QString dir = QMPlay2Core.getSettings().getString("screenshotPth");
-    quint16 num = 0;
-    for (const QString &f : QDir(dir).entryList({"QMPlay2_snap_?????" + ext}, QDir::Files, QDir::Name))
+    QString realUrl = playC.getUrl();
+    Functions::splitPrefixAndUrlIfHasPluginPrefix(realUrl, nullptr, &realUrl);
+    QString screenshotName;
+    if (realUrl.startsWith("file://"))
     {
-        const quint16 n = f.midRef(13, 5).toUShort();
-        if (n > num)
-            num = n;
+        screenshotName = Functions::fileName(realUrl) + "_QMPlay2_snapshot_" + Functions::timeToStr(videoFrame.ts(), false, true).replace(":", ".");
     }
-    img.save(dir + "/QMPlay2_snap_" + QString("%1").arg(++num, 5, 10, QChar('0')) + ext);
+    else
+    {
+        quint16 num = 0;
+        for (const QString &f : QDir(dir).entryList({"QMPlay2_snapshot_?????" + ext}, QDir::Files, QDir::Name))
+        {
+            const quint16 n = QStringView(f).mid(13, 5)MAYBE_TO_STRING.toUShort();
+            if (n > num)
+                num = n;
+        }
+        screenshotName = "QMPlay2_snap_" + QString("%1").arg(++num, 5, 10, QChar('0'));
+    }
+    screenshotName += ext;
+    if (img.save(dir + "/" + screenshotName))
+        playC.messageAndOSD(tr("Screenshot saved as: %1").arg(screenshotName));
 }
 void VideoThr::pause()
 {
     QMPlay2GUI.screenSaver->unInhibit(0);
+#ifdef Q_OS_WIN
+    setHighTimerResolution<false>();
+#endif
     writer->pause();
 }
