@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2020  Błażej Szczygieł
+    Copyright (C) 2010-2023  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -50,10 +50,6 @@
 
 #include <algorithm>
 
-#if defined(QMPLAY2_CPU_X86_32) && !defined(JS_SKIP_SSE2_CHECK)
-#    include <libavutil/cpu.h>
-#endif
-
 Q_LOGGING_CATEGORY(mb, "MediaBrowser")
 
 constexpr const char *g_mediaBrowserBaseUrl = "https://raw.githubusercontent.com/zaps166/QMPlay2OnlineContents/master/";
@@ -240,7 +236,7 @@ MediaBrowserPages::MediaBrowserPages() :
     layout->addWidget(m_currentPage, 0, 1, 1, 1);
     layout->addWidget(m_nextPage, 0, 2, 1, 1);
     layout->setSpacing(3);
-    layout->setMargin(0);
+    layout->setContentsMargins(0, 0, 0, 0);
 }
 MediaBrowserPages::~MediaBrowserPages()
 {}
@@ -501,14 +497,6 @@ void MediaBrowser::completionsReady()
 
 bool MediaBrowser::scanScripts()
 {
-#if defined(QMPLAY2_CPU_X86_32) && !defined(JS_SKIP_SSE2_CHECK)
-    if (!(QMPlay2CoreClass::getCPUFlags() & AV_CPU_FLAG_SSE2))
-    {
-        qCCritical(mb) << "SSE2 capable CPU is required";
-        return false;
-    }
-#endif
-
     const auto lastName = m_providersB->currentText();
 
     m_providersB->clear();
@@ -594,13 +582,12 @@ void MediaBrowser::downloadScripts(const QByteArray &jsonData)
     }();
 
     bool removed = false, downloading = false;
-    for (auto &&jsonVal : jsonArr)
+    for (const QJsonValue jsonVal : jsonArr)
     {
-        const auto jsonObj = jsonVal.toObject();
-        const auto name = jsonObj["Name"].toString();
-        const auto path = jsonObj["Path"].toString();
-        const auto version = jsonObj["Version"].toInt();
-        const auto apiVersion = jsonObj["ApiVersion"].toInt();
+        const auto name = jsonVal["Name"].toString();
+        const auto path = jsonVal["Path"].toString();
+        const auto version = jsonVal["Version"].toInt();
+        const auto apiVersion = jsonVal["ApiVersion"].toInt();
         if (name.isEmpty() || version <= 0 || apiVersion != g_mediaBrowserApiVersion)
             continue;
 
@@ -670,7 +657,7 @@ void MediaBrowser::providerChanged(int idx)
     if (m_mediaBrowser)
     {
         m_mediaBrowser->setCompleterListCallback(nullptr);
-        m_mediaBrowser->finalize();
+        m_mediaBrowser->finalize(true);
     }
 
     m_searchCB->blockSignals(true);
@@ -753,7 +740,7 @@ void MediaBrowser::search()
     if (m_imageReply)
         m_imageReply->deleteLater();
     if (m_mediaBrowser)
-        m_mediaBrowser->finalize();
+        m_mediaBrowser->finalize(false);
     m_resultsW->clear();
     if (!name.isEmpty())
     {
