@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2020  Błażej Szczygieł
+    Copyright (C) 2010-2023  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -22,15 +22,17 @@
 #include <QMPlay2Core.hpp>
 #include <Functions.hpp>
 #include <OggHelper.hpp>
-#include <CppUtils.hpp>
 #include <Settings.hpp>
 #include <Packet.hpp>
 
+#include <limits>
+
 extern "C"
 {
     #include <libavformat/avformat.h>
     #include <libavutil/replaygain.h>
     #include <libavutil/spherical.h>
+    #include <libavutil/display.h>
     #include <libavutil/pixdesc.h>
 }
 
@@ -78,7 +80,7 @@ static QByteArray getTag(AVDictionary *metadata, const char *key, const bool ded
     AVDictionaryEntry *avTag = av_dict_get(metadata, key, nullptr, 0);
     if (avTag && avTag->value)
     {
-        const QByteArray tag = QByteArray(avTag->value);
+        const QByteArray tag = Functions::textWithFallbackEncoding(QByteArray(avTag->value));
         if (deduplicate)
         {
             // Workaround for duplicated tags separated by ';'.
@@ -205,7 +207,7 @@ FormatContext::FormatContext(bool reconnectStreamed) :
     isPaused(false), fixMkvAss(false),
     isMetadataChanged(false),
     lastTime(0.0),
-    invalErrCount(0), errFromSeek(0),
+    m_retErrCount(0), errFromSeek(0),
     maybeHasFrame(false),
     artistWithTitle(true),
     stillImage(false),
@@ -219,7 +221,7 @@ FormatContext::~FormatContext()
         av_packet_free(&packet);
     }
     delete oggHelper;
-    for (StreamInfo *streamInfo : asConst(streamsInfo))
+    for (StreamInfo *streamInfo : qAsConst(streamsInfo))
         delete streamInfo;
 }
 
@@ -249,25 +251,24 @@ QList<ProgramInfo> FormatContext::getPrograms() const
     for (unsigned i = 0; i < formatCtx->nb_programs; ++i)
     {
         const AVProgram &program = *formatCtx->programs[i];
-        if (program.discard != AVDISCARD_ALL)
+        ProgramInfo programInfo(program.id);
+
+        const QString bitRateStr = getTag(program.metadata, "variant_bitrate", false);
+        if (!bitRateStr.isEmpty())
+            programInfo.bitrate = bitRateStr.toLongLong();
+
+        for (unsigned s = 0; s < program.nb_stream_indexes; ++s)
         {
-            ProgramInfo programInfo(program.program_num);
-            if (program.nb_stream_indexes)
+            const int ff_idx = program.stream_index[s];
+            const int idx = index_map[ff_idx];
+            if (idx > -1)
             {
-                for (unsigned s = 0; s < program.nb_stream_indexes; ++s)
-                {
-                    const int ff_idx = program.stream_index[s];
-                    const int idx = index_map[ff_idx];
-                    if (idx > -1)
-                    {
-                        const AVMediaType type = streams[ff_idx]->codecpar->codec_type;
-                        if (type != AVMEDIA_TYPE_UNKNOWN)
-                            programInfo.streams += {idx, type};
-                    }
-                }
+                const AVMediaType type = streams[ff_idx]->codecpar->codec_type;
+                if (type != AVMEDIA_TYPE_UNKNOWN)
+                    programInfo.streams += {idx, type};
             }
-            programs += programInfo;
         }
+        programs += programInfo;
     }
     return programs;
 }
@@ -505,6 +506,7 @@ bool FormatContext::seek(double pos, bool backward)
             currPos = pos;
             nextDts.fill(pos);
             isError = false;
+            m_retErrCount = 0;
         }
     }
     return isOk;
@@ -517,6 +519,13 @@ bool FormatContext::read(Packet &encoded, int &idx)
         return false;
     }
 
+    if (m_allDiscarded && !maybeHasFrame)
+    {
+        if (!isPaused)
+            pause();
+        return false;
+    }
+
     if (isPaused)
     {
         isPaused = false;
@@ -537,20 +546,25 @@ bool FormatContext::read(Packet &encoded, int &idx)
         errFromSeek = 0;
     }
 
-    if (ret == AVERROR_INVALIDDATA)
+    if (ret == AVERROR_INVALIDDATA || ret == AVERROR_EXIT)
     {
-        if (invalErrCount < 1000)
+        if (m_retErrCount < 1000)
         {
-            ++invalErrCount;
+            ++m_retErrCount;
             return true;
         }
         isError = true;
         return false;
     }
     else
-        invalErrCount = 0;
+    {
+        m_retErrCount = 0;
+    }
+
     if (ret == AVERROR(EAGAIN))
+    {
         return true;
+    }
     else if (ret)
     {
         isError = true;
@@ -590,7 +604,6 @@ bool FormatContext::read(Packet &encoded, int &idx)
         }
         encoded.setDuration(newDuration);
     }
-    streamsTS[ff_idx] = encoded.ts();
 
     if (isStreamed)
     {
@@ -610,9 +623,11 @@ bool FormatContext::read(Packet &encoded, int &idx)
         return false;
     }
 
-    // Generate DTS for key frames if DTS doesn't exist (workaround for some M3U8 seekable streams)
+    // Generate DTS for key frames if DTS doesn't exist (e.g. workaround for some M3U8 seekable streams or other streams)
     if (encoded.hasKeyFrame() && !encoded.hasDts())
         encoded.setDts(nextDts.at(ff_idx));
+
+    streamsTS[ff_idx] = encoded.ts();
     nextDts[ff_idx] = encoded.ts() + encoded.duration();
 
     currPos = encoded.ts();
@@ -637,6 +652,7 @@ bool FormatContext::open(const QString &_url, const QString &param)
         "ass",
         "tty", //txt files
         "srt",
+        "vplayer",
     };
 
     const QByteArray scheme = Functions::getUrlScheme(_url).toUtf8();
@@ -686,7 +702,7 @@ bool FormatContext::open(const QString &_url, const QString &param)
         if (scheme != "rtsp")
         {
             // It is needed for QMPlay2 schemes like "alsa://", "v4l2://", etc.
-            inputFmt = av_find_input_format(scheme);
+            inputFmt = const_cast<AVInputFormat *>(av_find_input_format(scheme));
             if (inputFmt)
                 url = _url.right(_url.length() - scheme.length() - 3);
         }
@@ -717,26 +733,61 @@ bool FormatContext::open(const QString &_url, const QString &param)
     // Useful, e.g. CUVID decoder needs valid PTS
     formatCtx->flags |= AVFMT_FLAG_GENPTS;
 
+    if (url.endsWith("sdp"))
+        av_dict_set(&options, "protocol_whitelist", "file,crypto,data,udp,rtp", 0);
+
     OpenFmtCtxThr *openThr = new OpenFmtCtxThr(formatCtx, url.toUtf8(), inputFmt, options, abortCtx);
     formatCtx = openThr->getFormatCtx();
     openThr->drop();
     if (!formatCtx || disabledDemuxers.contains(name()))
         return false;
 
-    if (name().startsWith("image2") || name().endsWith("_pipe"))
+    if (name() == "sdp")
+    {
+        isLocal = false;
+        formatCtx->flags |= AVFMT_FLAG_NOBUFFER;
+    }
+    else if (name().startsWith("image2") || name().endsWith("_pipe"))
     {
         if (!settings.getBool("StillImages"))
             return false;
         stillImage = true;
     }
-
-    if (name() == "mp3")
+    else if (name() == "mp3")
+    {
         formatCtx->flags |= AVFMT_FLAG_FAST_SEEK; //This should be set before "avformat_open_input", but seems to be working for MP3...
+    }
+
+    if (name() == "mpegts")
+    {
+        // Workaround to detect initial audio parameters for some TS files
+        formatCtx->probesize *= 2;
+    }
 
     if (avformat_find_stream_info(formatCtx, nullptr) < 0)
         return false;
 
-    isStreamed = !isLocal && formatCtx->duration <= 0; //QMPLAY2_NOPTS_VALUE is negative
+    // Determine the duration of WavPack if not known
+    if (isLocal && formatCtx->nb_streams == 1 && formatCtx->duration == AV_NOPTS_VALUE)
+    {
+        const auto stream = formatCtx->streams[0];
+        if (stream->codecpar->codec_id == AV_CODEC_ID_WAVPACK)
+        {
+            if (av_seek_frame(formatCtx, 0, std::numeric_limits<int64_t>::max(), AVSEEK_FLAG_BACKWARD) >= 0)
+            {
+                auto pkt = av_packet_alloc();
+                if (av_read_frame(formatCtx, pkt) == 0)
+                {
+                    if (pkt->dts != AV_NOPTS_VALUE)
+                        formatCtx->duration = av_rescale_q(pkt->dts + pkt->duration, stream->time_base, {1, AV_TIME_BASE});
+                }
+                av_packet_free(&pkt);
+            }
+            av_seek_frame(formatCtx, 0, 0, AVSEEK_FLAG_BACKWARD);
+        }
+    }
+
+    isStreamed = !isLocal && formatCtx->duration <= 0; // AV_NOPTS_VALUE is negative
 
 #ifdef QMPlay2_libavdevice
     forceCopy = name().contains("v4l2"); //Workaround for v4l2 - if many buffers are referenced demuxer doesn't produce proper timestamps (FFmpeg BUG?).
@@ -758,16 +809,28 @@ bool FormatContext::open(const QString &_url, const QString &param)
     streamsTS.resize(formatCtx->nb_streams);
     streamsOffset.resize(formatCtx->nb_streams);
     nextDts.resize(formatCtx->nb_streams);
+
+    int nVideoStreams = 0;
+    int nDefaultVideoStreams = 0;
     for (unsigned i = 0; i < formatCtx->nb_streams; ++i)
     {
         fixFontsAttachment(formatCtx->streams[i]);
         StreamInfo *streamInfo = getStreamInfo(formatCtx->streams[i]);
         if (!streamInfo)
+        {
             index_map[i] = -1;
+        }
         else
         {
             index_map[i] = streamsInfo.count();
             streamsInfo += streamInfo;
+
+            if (streamInfo->params->codec_type == AVMEDIA_TYPE_VIDEO)
+            {
+                if (streamInfo->is_default)
+                    ++nDefaultVideoStreams;
+                ++nVideoStreams;
+            }
         }
         if (!fixMkvAss && formatCtx->streams[i]->codecpar->codec_id == AV_CODEC_ID_ASS && !strncasecmp(formatCtx->iformat->name, "matroska", 8))
             fixMkvAss = true;
@@ -781,9 +844,72 @@ bool FormatContext::open(const QString &_url, const QString &param)
 
     isOneStreamOgg = (name() == "ogg" && streamsInfo.count() == 1); //Workaround for OGG network streams
 
-    if (isStreamed && streamsInfo.count() == 1 && streamsInfo.at(0)->codec_type == AVMEDIA_TYPE_SUBTITLE && formatCtx->pb && avio_size(formatCtx->pb) > 0)
+    if (isStreamed && streamsInfo.count() == 1 && streamsInfo.at(0)->params->codec_type == AVMEDIA_TYPE_SUBTITLE && formatCtx->pb && avio_size(formatCtx->pb) > 0)
         isStreamed = false; //Allow subtitles streams to be non-streamed if size is known
 
+    // Choose best quality streams for HLS or DASH
+    if (nVideoStreams > 1 && (nDefaultVideoStreams == 0 || nVideoStreams == nDefaultVideoStreams) && (name() == "hls" || name() == "dash"))
+    {
+        int streamInfoMaxBitRate = -1;
+
+        // Find stream with highest bitrate, unset default video streams
+        int64_t maxBitRate = 0;
+        for (int i = 0; i < streamsInfo.count(); ++i)
+        {
+            auto &streamInfo = streamsInfo.at(i);
+
+            if (streamInfo->params->codec_type != AVMEDIA_TYPE_VIDEO)
+                continue;
+
+            streamInfo->is_default = false;
+            if (streamInfo->params->bit_rate > maxBitRate)
+            {
+                maxBitRate = streamInfo->params->bit_rate;
+                streamInfoMaxBitRate = i;
+            }
+        }
+
+        if (streamInfoMaxBitRate > -1)
+        {
+            // Set best video stream as default
+            streamsInfo.at(streamInfoMaxBitRate)->is_default = true;
+
+            // Find first audio stream in the same program for the new default video stream and set it as default, too
+            for (unsigned i = 0; i < formatCtx->nb_programs; ++i)
+            {
+                const AVProgram &program = *formatCtx->programs[i];
+                int firstAudioIdx = -1;
+                bool found = false;
+                for (unsigned s = 0; s < program.nb_stream_indexes; ++s)
+                {
+                    const int ff_idx = program.stream_index[s];
+                    const int idx = index_map.at(ff_idx);
+                    if (streamInfoMaxBitRate == idx)
+                    {
+                        found = true;
+                    }
+                    if (firstAudioIdx == -1 && streams.at(ff_idx)->codecpar->codec_type == AVMEDIA_TYPE_AUDIO)
+                    {
+                        firstAudioIdx = idx;
+                    }
+                }
+                if (found)
+                {
+                    if (firstAudioIdx > -1)
+                    {
+                        for (auto &&streamInfo : qAsConst(streamsInfo))
+                        {
+                            if (streamInfo->params->codec_type == AVMEDIA_TYPE_AUDIO)
+                                streamInfo->is_default = false;
+                        }
+                        streamsInfo.at(firstAudioIdx)->is_default = true;
+                    }
+                    break;
+                }
+            }
+        }
+    }
+
     formatCtx->event_flags = 0;
 
     packet = av_packet_alloc();
@@ -801,6 +927,28 @@ void FormatContext::setStreamOffset(double offset)
         streamsOffset[i] = offset - streamsTS.at(i);
 }
 
+void FormatContext::selectStreams(const QSet<int> &selectedStreams)
+{
+    m_allDiscarded = true;
+    for (AVStream *stream : qAsConst(streams))
+    {
+        if (stream->codecpar->codec_type == AVMEDIA_TYPE_DATA || stream->codecpar->codec_type == AVMEDIA_TYPE_ATTACHMENT)
+        {
+            // Attachment is used once at "open()", data is not used at all.
+            stream->discard = AVDISCARD_ALL;
+            continue;
+        }
+
+        const int idx = index_map.at(stream->index);
+        stream->discard = (idx >= 0 && selectedStreams.contains(idx))
+            ? AVDISCARD_DEFAULT
+            : AVDISCARD_ALL
+        ;
+        if (stream->discard == AVDISCARD_DEFAULT)
+            m_allDiscarded = false;
+    }
+}
+
 /**/
 
 AVDictionary *FormatContext::getMetadata() const
@@ -815,15 +963,20 @@ StreamInfo *FormatContext::getStreamInfo(AVStream *stream) const
     StreamInfo *streamInfo = new StreamInfo(stream->codecpar);
 
     streamInfo->must_decode = true;
-    if (const AVCodecDescriptor *codecDescr = avcodec_descriptor_get(streamInfo->codec_id))
+    if (const AVCodecDescriptor *codecDescr = avcodec_descriptor_get(streamInfo->params->codec_id))
     {
         if (codecDescr->props & AV_CODEC_PROP_TEXT_SUB)
-            streamInfo->must_decode = false;
+        {
+            if (codecDescr->id == AV_CODEC_ID_MOV_TEXT)
+                streamInfo->decode_to_ass = true;
+            else
+                streamInfo->must_decode = false;
+        }
 
         if (streamInfo->codec_name.isEmpty())
             streamInfo->codec_name = codecDescr->name;
     }
-    else if (streamInfo->codec_type == AVMEDIA_TYPE_SUBTITLE)
+    else if (streamInfo->params->codec_type == AVMEDIA_TYPE_SUBTITLE)
     {
         streamInfo->must_decode = false;
     }
@@ -831,7 +984,7 @@ StreamInfo *FormatContext::getStreamInfo(AVStream *stream) const
     streamInfo->is_default = stream->disposition & AV_DISPOSITION_DEFAULT;
     streamInfo->time_base = stream->time_base;
 
-    if (streamInfo->codec_type != AVMEDIA_TYPE_ATTACHMENT)
+    if (streamInfo->params->codec_type != AVMEDIA_TYPE_ATTACHMENT)
     {
         QString value;
         if (streamsInfo.count() > 1)
@@ -849,22 +1002,51 @@ StreamInfo *FormatContext::getStreamInfo(AVStream *stream) const
         }
         if (!(value = getTag(stream->metadata, "language", false)).isEmpty() && value != "und")
             streamInfo->other_info += {QString::number(QMPLAY2_TAG_LANGUAGE), value};
+        if (streamInfo->params->codec_type == AVMEDIA_TYPE_VIDEO)
+        {
+            if (!(value = getTag(stream->metadata, "variant_bitrate", false)).isEmpty())
+            {
+                bool ok = false;
+                int64_t bitRate = value.toLongLong(&ok);
+                if (ok && bitRate > 0)
+                    streamInfo->params->bit_rate = bitRate;
+            }
+        }
     }
 
-    switch (streamInfo->codec_type)
+    switch (streamInfo->params->codec_type)
     {
         case AVMEDIA_TYPE_VIDEO:
         {
-            if (stream->sample_aspect_ratio.num)
-                streamInfo->sample_aspect_ratio = stream->sample_aspect_ratio;
+            if (stream->sample_aspect_ratio.num && av_cmp_q(streamInfo->params->sample_aspect_ratio, stream->sample_aspect_ratio) != 0)
+            {
+                streamInfo->params->sample_aspect_ratio = stream->sample_aspect_ratio;
+                streamInfo->custom_sar = true;
+            }
 
             if (!stillImage)
                 streamInfo->fps = stream->r_frame_rate;
 
-            bool ok = false;
-            const double rotation = getTag(stream->metadata, "rotate", false).toDouble(&ok);
-            if (ok)
-                streamInfo->rotation = rotation;
+            if (void *sideData = av_stream_get_side_data(stream, AV_PKT_DATA_DISPLAYMATRIX, nullptr))
+            {
+                const auto rotation = av_display_rotation_get(reinterpret_cast<const int32_t *>(sideData));
+                if (!qIsNaN(rotation))
+                {
+                    switch(qRound(rotation))
+                    {
+                        case -180:
+                        case 180:
+                            streamInfo->rotation = 180.0;
+                            break;
+                        case -90:
+                            streamInfo->rotation = 90.0;
+                            break;
+                        case 90:
+                            streamInfo->rotation = 270.0;
+                            break;
+                    }
+                }
+            }
 
             if (void *sideData = av_stream_get_side_data(stream, AV_PKT_DATA_SPHERICAL, nullptr))
             {
@@ -876,7 +1058,7 @@ StreamInfo *FormatContext::getStreamInfo(AVStream *stream) const
         case AVMEDIA_TYPE_ATTACHMENT:
         {
             streamInfo->title = getTag(stream->metadata, "filename", false);
-            switch (streamInfo->codec_id)
+            switch (streamInfo->params->codec_id)
             {
                 case AV_CODEC_ID_TTF:
                     streamInfo->codec_name = "TTF";
