@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2020  Błażej Szczygieł
+    Copyright (C) 2010-2023  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -21,7 +21,6 @@
 #include <VideoFilters.hpp>
 #include <GPUInstance.hpp>
 #include <Functions.hpp>
-#include <CppUtils.hpp>
 #ifdef USE_QML
 #   include <CommonJS.hpp>
 #endif
@@ -30,6 +29,7 @@
 #include <Module.hpp>
 
 #include <QLoggingCategory>
+#include <QStandardPaths>
 #include <QApplication>
 #include <QLibraryInfo>
 #include <QTranslator>
@@ -44,9 +44,10 @@
     #include <windows.h>
     #include <powrprof.h>
 #elif defined Q_OS_MACOS
-    #include <QStandardPaths>
-#elif !defined Q_OS_ANDROID
-    #include <QProcess>
+    #include <QOperatingSystemVersion>
+#elif !defined Q_OS_ANDROID && !defined Q_OS_HAIKU
+    #include <QDBusConnection>
+    #include <QDBusInterface>
 #endif
 
 #include <cstdarg>
@@ -59,16 +60,34 @@ extern "C"
     #include <libavutil/log.h>
 }
 
-#if !defined(Q_OS_ANDROID) && !defined(Q_OS_WIN) && !defined(Q_OS_MACOS)
-static QStringList getDBusSuspendArgs(const QString &method, const QString &name, const QString &pathSuffix = QString())
+#if defined DBUS_SUSPEND
+class DBusSuspend
 {
-    return {
-        "--system",
-        "--print-reply",
-        "--dest=org.freedesktop." + name,
-        "/org/freedesktop/" + name + pathSuffix,
-        "org.freedesktop." + name + ".Manager." + method
-     };
+public:
+    DBusSuspend(const QDBusConnection &connection, const QString &name, const QString &pathSuffix = QString())
+        : m_suspendIface(
+              "org.freedesktop." + name,
+              "/org/freedesktop/" + name + pathSuffix,
+              "org.freedesktop." + name + ".Manager",
+              connection
+          )
+    {}
+
+    bool canSuspend()
+    {
+        if (m_suspendIface.isValid())
+            return (m_suspendIface.call("CanSuspend").arguments().at(0).toString().toLower() == "yes");
+        return false;
+    }
+
+    void suspend()
+    {
+        if (m_suspendIface.isValid())
+            m_suspendIface.call("Suspend", QVariant::fromValue(true));
+    }
+
+private:
+    QDBusInterface m_suspendIface;
 };
 #endif
 
@@ -174,19 +193,37 @@ QString QMPlay2CoreClass::getLibDir()
 
 QString QMPlay2CoreClass::getLongFromShortLanguage(const QString &lng)
 {
-    const QString lang = QLocale::languageToString(QLocale(lng).language());
-    return lang == "C" ? lng : lang;
+    const QLocale locale(lng);
+    QString lang = locale.nativeLanguageName();
+    if (lng.contains(QChar('_')))
+        lang += " (" + locale.nativeCountryName() + ")";
+    return lang.isEmpty() ? lng : lang;
 }
 
 int QMPlay2CoreClass::getCPUFlags()
 {
     return av_get_cpu_flags();
 }
+int QMPlay2CoreClass::getCPUMaxAlign()
+{
+    return av_cpu_max_align();
+}
 
 #ifdef USE_OPENGL
 bool QMPlay2CoreClass::isGlOnWindowForced()
 {
-    static bool forced = (QApplication::platformName().startsWith("wayland") || QApplication::platformName() == "android");
+    static bool forced = [] {
+#ifdef Q_OS_MACOS
+#   if defined(MAC_OS_X_VERSION_10_14) && (MAC_OS_X_VERSION_MAX_ALLOWED >= MAC_OS_X_VERSION_10_14)
+        return (QOperatingSystemVersion::current() <= QOperatingSystemVersion::MacOSHighSierra);
+#   else
+#       warning "Compiling OpenGL with SDK older than 10.14, native OpenGL window is disabled!"
+        return true;
+#   endif
+#else
+        return (QApplication::platformName() == "android");
+#endif
+    }();
     return forced;
 }
 #endif
@@ -201,15 +238,27 @@ void QMPlay2CoreClass::init(bool loadModules, bool modulesInSubdirs, const QStri
     langDir = shareDir + "lang/";
 
     if (Version::isPortable())
+    {
         settingsDir = QCoreApplication::applicationDirPath() + "/settings/";
+    }
     else
     {
-#if defined(Q_OS_WIN)
+#if defined(Q_OS_WIN) || defined(Q_OS_HAIKU)
         settingsDir = QFileInfo(QSettings(QSettings::IniFormat, QSettings::UserScope, QString()).fileName()).absolutePath() + "/QMPlay2/";
 #elif defined(Q_OS_MACOS)
-        settingsDir = Functions::cleanPath(QStandardPaths::standardLocations(QStandardPaths::DataLocation).value(0, settingsDir));
+        settingsDir = Functions::cleanPath(QStandardPaths::standardLocations(QStandardPaths::AppDataLocation).value(0));
 #else
-        settingsDir = QDir::homePath() + "/.qmplay2/";
+        const QString configSettingsDir(QStandardPaths::standardLocations(QStandardPaths::ConfigLocation).value(0) + "/QMPlay2/");
+        const QString homeSettingsDir(QDir::homePath() + "/.qmplay2/");
+        if (QFileInfo(configSettingsDir).isDir() || !QFileInfo(homeSettingsDir).isDir())
+        {
+            settingsDir = configSettingsDir;
+        }
+        else
+        {
+            // Backward compatibility
+            settingsDir = homeSettingsDir;
+        }
 #endif
     }
     QDir(settingsDir).mkpath(".");
@@ -259,15 +308,12 @@ void QMPlay2CoreClass::init(bool loadModules, bool modulesInSubdirs, const QStri
     QCoreApplication::installTranslator(qtTranslator);
     setLanguage();
 
-#ifdef Q_OS_WIN
-    timeBeginPeriod(1); //Set the timer for 1ms resolution (for Sleep ())
+#ifdef QT_DEBUG
+    av_log_set_level(AV_LOG_INFO);
+#else
+    av_log_set_level(AV_LOG_WARNING);
 #endif
-
-    av_log_set_level(AV_LOG_ERROR);
     av_log_set_callback(avQMPlay2LogHandler);
-#if LIBAVCODEC_VERSION_INT < AV_VERSION_INT(58, 9, 100)
-    av_register_all();
-#endif
     avformat_network_init();
 
 #if defined(USE_VULKAN) && !defined(Q_OS_ANDROID)
@@ -300,7 +346,7 @@ void QMPlay2CoreClass::init(bool loadModules, bool modulesInSubdirs, const QStri
         }
 
         QStringList pluginsName;
-        for (const QFileInfo &fInfo : asConst(pluginsList))
+        for (const QFileInfo &fInfo : qAsConst(pluginsList))
         {
             if (QLibrary::isLibrary(fInfo.filePath()))
             {
@@ -373,7 +419,7 @@ void QMPlay2CoreClass::quit()
 {
     if (settingsDir.isEmpty())
         return;
-    for (Module *pluginInstance : asConst(pluginsInstance))
+    for (Module *pluginInstance : qAsConst(pluginsInstance))
         delete pluginInstance;
     pluginsInstance.clear();
     videoFilters.clear();
@@ -381,9 +427,6 @@ void QMPlay2CoreClass::quit()
     shareDir.clear();
     langDir.clear();
     avformat_network_deinit();
-#ifdef Q_OS_WIN
-    timeEndPeriod(1);
-#endif
     QCoreApplication::removeTranslator(qtTranslator);
     QCoreApplication::removeTranslator(translator);
     delete qtTranslator;
@@ -401,20 +444,14 @@ bool QMPlay2CoreClass::canSuspend()
 #if defined Q_OS_WIN || defined Q_OS_MACOS
     m_suspend = 1;
     return true;
-#elif !defined Q_OS_ANDROID
-    auto checkSuspendDBus = [](const QStringList &args) {
-        QProcess p;
-        p.start("dbus-send", args);
-        if (!p.waitForStarted() || !p.waitForFinished())
-            return false;
-        return (p.readAllStandardOutput().split('\n').value(1).simplified().replace("\"", "").split(' ').value(1).toLower() == "yes");
-    };
-    if (checkSuspendDBus(getDBusSuspendArgs("CanSuspend", "login1")))
+#elif defined DBUS_SUSPEND
+    auto connection = QDBusConnection::systemBus();
+    if (DBusSuspend(connection, "login1").canSuspend())
     {
         m_suspend = 1;
         return true;
     }
-    if (checkSuspendDBus(getDBusSuspendArgs("CanSuspend", "ConsoleKit", "/Manager")))
+    if (DBusSuspend(connection, "ConsoleKit", "/Manager").canSuspend())
     {
         m_suspend = 2;
         return true;
@@ -431,25 +468,19 @@ void QMPlay2CoreClass::suspend()
     SetSuspendState(false, false, false);
 #elif defined Q_OS_MACOS
     Q_UNUSED(system("pmset sleepnow > /dev/null 2>&1 &"));
-#elif !defined Q_OS_ANDROID
-    QStringList args;
+#elif defined DBUS_SUSPEND
+    auto connection = QDBusConnection::systemBus();
     switch (m_suspend)
     {
         case 1:
-            args = getDBusSuspendArgs("Suspend", "login1");
+            DBusSuspend(connection, "login1").suspend();
             break;
         case 2:
-            args = getDBusSuspendArgs("Suspend", "ConsoleKit", "/Manager");
+            DBusSuspend(connection, "ConsoleKit", "/Manager").suspend();
             break;
         default:
             return;
     }
-    args += "boolean:true";
-
-    QProcess p;
-    p.start("dbus-send", args);
-    if (p.waitForStarted())
-        p.waitForFinished();
 #endif
 }
 
@@ -460,10 +491,15 @@ QStringList QMPlay2CoreClass::getModules(const QString &type, int typeLen) const
     if (type == "videoWriters")
         defaultModules << "OpenGL 2" << "XVideo";
     else if (type == "audioWriters")
-        defaultModules << "PulseAudio" << "ALSA";
+        defaultModules << "PipeWire" << "PulseAudio" << "ALSA";
 #elif defined Q_OS_WIN
     if (type == "videoWriters")
         defaultModules << "OpenGL 2" << "DirectDraw";
+#elif defined Q_OS_HAIKU
+    if ( type == "videoWriters" )
+        defaultModules << "QPainter";
+    else if ( type == "audioWriters" )
+        defaultModules << "MediaKit";
 #endif
     if (type == "decoders")
         defaultModules << "FFmpeg Decoder";
@@ -551,16 +587,21 @@ QStringList QMPlay2CoreClass::getLanguages() const
 }
 void QMPlay2CoreClass::setLanguage()
 {
-    QString systemLang = QLocale::system().name();
-    const int idx = systemLang.indexOf('_');
-    if (idx > -1)
-        systemLang.remove(idx, systemLang.size() - idx);
-    lang = settings->getString("Language", systemLang);
+    lang = settings->getString("Language");
     if (lang.isEmpty())
-        lang = systemLang;
-    if (!translator->load(lang, langDir))
+        lang = QLocale::system().name();
+    if (translator->load(lang, langDir))
+    {
+#if QT_VERSION >= QT_VERSION_CHECK(5, 15, 0)
+        lang = QFileInfo(translator->filePath()).baseName();
+#endif
+    }
+    else
+    {
         lang = "en";
-    qtTranslator->load("qtbase_" + lang, QLibraryInfo::location(QLibraryInfo::TranslationsPath));
+    }
+    const bool loaded = qtTranslator->load("qtbase_" + lang, QLibraryInfo::location(QLibraryInfo::TranslationsPath));
+    Q_UNUSED(loaded)
 }
 
 void QMPlay2CoreClass::addVideoDeintMethod(QWidget *w)
@@ -628,6 +669,16 @@ QString QMPlay2CoreClass::getNameForUrl(const QString &url)
     return getDataFromHash(url, namesForUrl);
 }
 
+void QMPlay2CoreClass::addDescriptionForUrl(const QString &url, const QString &description, const bool removeAfterUse)
+{
+    if (!url.isEmpty())
+        setDataToHash(url, description.toUtf8(), removeAfterUse, descriptionsForUrl);
+}
+QString QMPlay2CoreClass::getDescriptionForUrl(const QString &url)
+{
+    return getDataFromHash(url, descriptionsForUrl);
+}
+
 void QMPlay2CoreClass::loadPlaylistGroup(const QString &name, const QMPlay2CoreClass::GroupEntries &entries, bool enqueue)
 {
     if (!entries.isEmpty())
@@ -660,12 +711,22 @@ QMPlay2CoreClass::Renderer QMPlay2CoreClass::renderer() const
 bool QMPlay2CoreClass::isGlOnWindow() const
 {
 #ifdef USE_OPENGL
-    return (isGlOnWindowForced() || settings->getBool("OpenGL/OnWindow"));
+    return (renderer() == QMPlay2CoreClass::Renderer::OpenGL) && (isGlOnWindowForced() || settings->getBool("OpenGL/OnWindow"));
 #else
     return false;
 #endif
 }
 
+void QMPlay2CoreClass::registerProcessWheelEventFn(const ProcessWheelEventFn &fn)
+{
+    m_processWheelEventFn = fn;
+}
+void QMPlay2CoreClass::processWheelEvent(QWheelEvent *e)
+{
+    if (m_processWheelEventFn)
+        m_processWheelEventFn(e);
+}
+
 void QMPlay2CoreClass::restoreCursorSlot()
 {
     QApplication::restoreOverrideCursor();
