@@ -1,6 +1,6 @@
 /*
     QMPlay2 is a video and audio player.
-    Copyright (C) 2010-2020  Błażej Szczygieł
+    Copyright (C) 2010-2023  Błażej Szczygieł
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU Lesser General Public License as published
@@ -60,32 +60,36 @@ QString StreamInfo::getTagName(const QString &tag)
 
 StreamInfo::StreamInfo()
 {
-    memset(static_cast<AVCodecParameters *>(this), 0, sizeof(AVCodecParameters));
-    format = -1;
+    params = avcodec_parameters_alloc();
+    params->format = -1;
     resetSAR();
 }
 StreamInfo::StreamInfo(AVCodecParameters *codecpar)
     : StreamInfo()
 {
-    avcodec_parameters_copy(static_cast<AVCodecParameters *>(this), codecpar);
+    avcodec_parameters_copy(params, codecpar);
 
-    if (const AVCodec *codec = avcodec_find_decoder(codec_id))
+    if (const AVCodec *codec = avcodec_find_decoder(params->codec_id))
         codec_name = codec->name;
 
-    if (sample_aspect_ratio.num == 0)
+    if (params->sample_aspect_ratio.num == 0)
         resetSAR();
 }
 StreamInfo::StreamInfo(quint32 sampleRateArg, quint8 channelsArg)
     : StreamInfo()
 {
-    codec_type = AVMEDIA_TYPE_AUDIO;
-    sample_rate = sampleRateArg;
-    channels = channelsArg;
+    params->codec_type = AVMEDIA_TYPE_AUDIO;
+    params->sample_rate = sampleRateArg;
+    params->channels = channelsArg;
+}
+StreamInfo::~StreamInfo()
+{
+    avcodec_parameters_free(&params);
 }
 
 QByteArray StreamInfo::getFormatName() const
 {
-    switch (codec_type)
+    switch (params->codec_type)
     {
         case AVMEDIA_TYPE_AUDIO:
             return av_get_sample_fmt_name(sampleFormat());
@@ -98,10 +102,35 @@ QByteArray StreamInfo::getFormatName() const
 }
 void StreamInfo::setFormat(int newFormat)
 {
-    format = newFormat;
+    params->format = newFormat;
+}
+
+QByteArray StreamInfo::getColorRangeName() const
+{
+    if (params->color_range != AVCOL_RANGE_UNSPECIFIED)
+        return av_color_range_name(params->color_range);
+    return QByteArray();
+}
+QByteArray StreamInfo::getColorPrimariesName() const
+{
+    if (params->color_primaries != AVCOL_PRI_UNSPECIFIED)
+        return av_color_primaries_name(params->color_primaries);
+    return QByteArray();
+}
+QByteArray StreamInfo::getColorTrcName() const
+{
+    if (params->color_trc != AVCOL_TRC_UNSPECIFIED)
+        return av_color_transfer_name(params->color_trc);
+    return QByteArray();
+}
+QByteArray StreamInfo::getColorSpaceName() const
+{
+    if (params->color_space != AVCOL_SPC_UNSPECIFIED)
+        return av_color_space_name(params->color_space);
+    return QByteArray();
 }
 
 inline void StreamInfo::resetSAR()
 {
-    sample_aspect_ratio.num = sample_aspect_ratio.den = 1;
+    params->sample_aspect_ratio.num = params->sample_aspect_ratio.den = 1;
 }
