@@ -1,35 +1,36 @@
+// SPDX-License-Identifier: MIT
 /*
-    QmVk - simple Vulkan library created for QMPlay2
-    Copyright (C) 2020  Błażej Szczygieł
-
-    This program is free software: you can redistribute it and/or modify
-    it under the terms of the GNU Lesser General Public License as published
-    by the Free Software Foundation, either version 3 of the License, or
-    (at your option) any later version.
-
-    This program is distributed in the hope that it will be useful,
-    but WITHOUT ANY WARRANTY; without even the implied warranty of
-    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-    GNU Lesser General Public License for more details.
-
-    You should have received a copy of the GNU Lesser General Public License
-    along with this program.  If not, see <http://www.gnu.org/licenses/>.
+   QmVk - simple Vulkan library created for QMPlay2
+   Copyright (C) 2020-2023 Błażej Szczygieł
 */
 
 #include "Image.hpp"
 #include "PhysicalDevice.hpp"
 #include "Device.hpp"
 #include "MemoryPropertyFlags.hpp"
-#include "DescriptorInfo.hpp"
 #include "CommandBuffer.hpp"
-#include "Buffer.hpp"
-#include "BufferView.hpp"
+#ifdef QMVK_USE_IMAGE_BUFFER_VIEW
+#   include "Buffer.hpp"
+#   include "BufferView.hpp"
+#endif
 
 #include <functional>
 #include <cmath>
 
 namespace QmVk {
 
+bool Image::checkFormatSampledImage(
+    const shared_ptr<PhysicalDevice> &physicalDevice,
+    vk::Format fmt,
+    bool linear)
+{
+    const auto &formatProperties = physicalDevice->getFormatPropertiesCached(fmt);
+    const auto flags = vk::FormatFeatureFlagBits::eSampledImage | vk::FormatFeatureFlagBits::eSampledImageFilterLinear;
+    if (linear)
+        return static_cast<bool>(formatProperties.linearTilingFeatures & flags);
+    return static_cast<bool>(formatProperties.optimalTilingFeatures & flags);
+}
+
 uint32_t Image::getNumPlanes(vk::Format format)
 {
     switch (format)
@@ -69,9 +70,10 @@ vk::ExternalMemoryProperties Image::getExternalMemoryProperties(
         : vk::ImageTiling::eOptimal
     ;
     imageFormatInfo.usage =
-        vk::ImageUsageFlagBits::eTransferSrc |
-        vk::ImageUsageFlagBits::eSampled
+        vk::ImageUsageFlagBits::eTransferSrc
     ;
+    if (checkFormatSampledImage(physicalDevice, realFmt, linear))
+        imageFormatInfo.usage |= vk::ImageUsageFlagBits::eSampled;
     imageFormatInfo.pNext = &externalImageFormatInfo;
 
     return physicalDevice->getImageFormatProperties2KHR<
@@ -103,15 +105,17 @@ shared_ptr<Image> Image::createOptimal(
         exportMemoryTypes,
         Priv()
     );
-    image->init(true, heap);
+    image->init(MemoryPropertyPreset::PreferNoHostAccess, heap);
     return image;
 }
 shared_ptr<Image> Image::createLinear(
     const shared_ptr<Device> &device,
     const vk::Extent2D &size,
     vk::Format fmt,
+    MemoryPropertyPreset memoryPropertyPreset,
     uint32_t paddingHeight,
-    bool deviceLocal,
+    bool useMipMaps,
+    bool storage,
     vk::ExternalMemoryHandleTypeFlags exportMemoryTypes,
     uint32_t heap)
 {
@@ -121,13 +125,13 @@ shared_ptr<Image> Image::createLinear(
         fmt,
         paddingHeight,
         true,
-        false,
-        false,
+        useMipMaps,
+        storage,
         false,
         exportMemoryTypes,
         Priv()
     );
-    image->init(deviceLocal, heap);
+    image->init(memoryPropertyPreset, heap);
     return image;
 }
 
@@ -135,7 +139,9 @@ shared_ptr<Image> Image::createExternalImport(
     const shared_ptr<Device> &device,
     const vk::Extent2D &size,
     vk::Format fmt,
-    bool linear)
+    bool linear,
+    vk::ExternalMemoryHandleTypeFlags exportMemoryTypes,
+    ImageCreateInfoCallback imageCreateInfoCallback)
 {
     auto image = make_shared<Image>(
         device,
@@ -146,10 +152,10 @@ shared_ptr<Image> Image::createExternalImport(
         false,
         false,
         true,
-        vk::ExternalMemoryHandleTypeFlags(),
+        exportMemoryTypes,
         Priv()
     );
-    image->init();
+    image->init(MemoryPropertyPreset::PreferNoHostAccess, ~0u, imageCreateInfoCallback);
     return image;
 }
 
@@ -179,7 +185,10 @@ Image::~Image()
     unmap();
 }
 
-void Image::init(bool deviceLocal, uint32_t heap)
+void Image::init(
+    MemoryPropertyPreset memoryPropertyPreset,
+    uint32_t heap,
+    ImageCreateInfoCallback imageCreateInfoCallback)
 {
     if (!m_externalImport && m_useMipMaps)
     {
@@ -265,10 +274,23 @@ void Image::init(bool deviceLocal, uint32_t heap)
             break;
     }
 
+    auto checkSampledImage = [this] {
+        for (uint32_t i = 0; i < m_numPlanes; ++i)
+        {
+            if (!checkFormatSampledImage(m_physicalDevice, m_formats[i], m_linear))
+                return false;
+        }
+        return true;
+    };
+
     vk::ImageUsageFlags imageUsageFlags =
-        vk::ImageUsageFlagBits::eTransferSrc |
-        vk::ImageUsageFlagBits::eSampled
+        vk::ImageUsageFlagBits::eTransferSrc
     ;
+    if (checkSampledImage())
+    {
+        imageUsageFlags |= vk::ImageUsageFlagBits::eSampled;
+        m_sampled = true;
+    }
     if (!m_externalImport)
         imageUsageFlags |= vk::ImageUsageFlagBits::eTransferDst;
     if (m_storage)
@@ -288,10 +310,7 @@ void Image::init(bool deviceLocal, uint32_t heap)
             : vk::ImageTiling::eOptimal
         ;
         imageCreateInfo.usage = imageUsageFlags;
-        imageCreateInfo.initialLayout = m_externalImport
-            ? vk::ImageLayout::ePreinitialized
-            : vk::ImageLayout::eUndefined
-        ;
+        imageCreateInfo.initialLayout = vk::ImageLayout::eUndefined;
 
         vk::ExternalMemoryImageCreateInfo externalMemoryImageCreateInfo;
         if (m_exportMemoryTypes)
@@ -300,12 +319,15 @@ void Image::init(bool deviceLocal, uint32_t heap)
             imageCreateInfo.pNext = &externalMemoryImageCreateInfo;
         }
 
+        if (imageCreateInfoCallback)
+            imageCreateInfoCallback(i, imageCreateInfo);
+
         m_images[i] = m_device->createImageUnique(imageCreateInfo);
     }
 
-    allocateAndBindMemory(deviceLocal, heap);
+    allocateAndBindMemory(memoryPropertyPreset, heap);
 }
-void Image::allocateAndBindMemory(bool deviceLocal, uint32_t heap)
+void Image::allocateAndBindMemory(MemoryPropertyPreset memoryPropertyPreset, uint32_t heap)
 {
     vector<vk::DeviceSize> memoryOffsets(m_numPlanes);
 
@@ -326,6 +348,7 @@ void Image::allocateAndBindMemory(bool deviceLocal, uint32_t heap)
 
         const auto memoryRequirements = m_device->getImageMemoryRequirements(*m_images[i]);
         const auto memoryRequirementsSize = aligned(memoryRequirements.size + paddingBytes, memoryRequirements.alignment);
+
         m_memoryRequirements.size += memoryRequirementsSize;
         m_memoryRequirements.alignment = max(m_memoryRequirements.alignment, memoryRequirements.alignment);
         m_memoryRequirements.memoryTypeBits |= memoryRequirements.memoryTypeBits;
@@ -338,24 +361,81 @@ void Image::allocateAndBindMemory(bool deviceLocal, uint32_t heap)
     if (m_externalImport)
         return;
 
+#ifdef QMVK_USE_IMAGE_BUFFER_VIEW
+        vk::BufferCreateInfo bufferCreateInfo;
+        bufferCreateInfo.usage = vk::BufferUsageFlagBits::eUniformTexelBuffer | vk::BufferUsageFlagBits::eStorageTexelBuffer;
+        bufferCreateInfo.size = m_memoryRequirements.size;
+        m_uniqueBuffer = m_device->createBufferUnique(bufferCreateInfo);
+
+        m_memoryRequirements.alignment = max(
+            m_memoryRequirements.alignment,
+            m_device->getBufferMemoryRequirements(*m_uniqueBuffer).alignment
+        );
+        m_memoryRequirements.size = aligned(m_memoryRequirements.size, m_memoryRequirements.alignment);
+#endif
+
     MemoryPropertyFlags memoryPropertyFlags;
-    if (deviceLocal)
+    switch (memoryPropertyPreset)
     {
-        memoryPropertyFlags.required =
-            vk::MemoryPropertyFlagBits::eDeviceLocal
-        ;
-        memoryPropertyFlags.notWanted =
-            vk::MemoryPropertyFlagBits::eHostVisible |
-            vk::MemoryPropertyFlagBits::eHostCached
-        ;
-    }
-    else
-    {
-        memoryPropertyFlags.required =
-            vk::MemoryPropertyFlagBits::eHostVisible |
-            vk::MemoryPropertyFlagBits::eHostCoherent |
-            vk::MemoryPropertyFlagBits::eHostCached
-        ;
+        case MemoryPropertyPreset::PreferNoHostAccess:
+            memoryPropertyFlags.required =
+                vk::MemoryPropertyFlagBits::eDeviceLocal
+            ;
+            memoryPropertyFlags.notWanted =
+                vk::MemoryPropertyFlagBits::eHostVisible
+            ;
+            break;
+        case MemoryPropertyPreset::PreferCachedOrNoHostAccess:
+            memoryPropertyFlags.required =
+                vk::MemoryPropertyFlagBits::eDeviceLocal
+            ;
+            memoryPropertyFlags.optional =
+                vk::MemoryPropertyFlagBits::eHostVisible |
+                vk::MemoryPropertyFlagBits::eHostCoherent |
+                vk::MemoryPropertyFlagBits::eHostCached
+            ;
+            break;
+        case MemoryPropertyPreset::PreferHostAccess:
+            memoryPropertyFlags.required =
+                vk::MemoryPropertyFlagBits::eDeviceLocal
+            ;
+            memoryPropertyFlags.optional =
+                vk::MemoryPropertyFlagBits::eHostVisible |
+                vk::MemoryPropertyFlagBits::eHostCoherent |
+                vk::MemoryPropertyFlagBits::eHostCached
+            ;
+            memoryPropertyFlags.optionalFallback =
+                vk::MemoryPropertyFlagBits::eHostVisible |
+                vk::MemoryPropertyFlagBits::eHostCoherent
+            ;
+            break;
+        case MemoryPropertyPreset::PreferCachedHostOnly:
+            memoryPropertyFlags.required =
+                vk::MemoryPropertyFlagBits::eHostVisible |
+                vk::MemoryPropertyFlagBits::eHostCoherent
+            ;
+            memoryPropertyFlags.optional =
+                vk::MemoryPropertyFlagBits::eHostCached
+            ;
+            if (m_physicalDevice->hasFullHostVisibleDeviceLocal())
+            {
+                memoryPropertyFlags.optionalFallback =
+                    vk::MemoryPropertyFlagBits::eDeviceLocal
+                ;
+            }
+            break;
+        case MemoryPropertyPreset::PreferHostOnly:
+            memoryPropertyFlags.required =
+                vk::MemoryPropertyFlagBits::eHostVisible |
+                vk::MemoryPropertyFlagBits::eHostCoherent
+            ;
+            memoryPropertyFlags.optional =
+                vk::MemoryPropertyFlagBits::eDeviceLocal
+            ;
+            memoryPropertyFlags.optionalFallback =
+                vk::MemoryPropertyFlagBits::eHostCached
+            ;
+            break;
     }
     memoryPropertyFlags.heap = heap;
     allocateMemory(memoryPropertyFlags);
@@ -371,7 +451,7 @@ void Image::finishImport(const vector<vk::DeviceSize> &offsets, vk::DeviceSize g
     {
         m_device->bindImageMemory(
             *m_images[i],
-            deviceMemory(min<uint32_t>(i, m_deviceMemory.size() - 1)),
+            deviceMemory(min<uint32_t>(i, deviceMemoryCount() - 1)),
             offsets[i] + globalOffset
         );
     }
@@ -380,6 +460,9 @@ void Image::finishImport(const vector<vk::DeviceSize> &offsets, vk::DeviceSize g
 
 void Image::createImageViews()
 {
+    if (!m_storage && !m_sampled)
+        return;
+
     for (uint32_t i = 0; i < m_numPlanes; ++i)
     {
         vk::ImageViewCreateInfo imageViewCreateInfo;
@@ -391,19 +474,26 @@ void Image::createImageViews()
     }
 }
 
+#ifdef QMVK_USE_IMAGE_BUFFER_VIEW
 shared_ptr<BufferView> Image::bufferView(uint32_t plane)
 {
     if (m_bufferViews.empty())
     {
+        if (!m_uniqueBuffer)
+            return nullptr;
+
+        auto buffer = Buffer::createFromDeviceMemory(
+            m_device,
+            memorySize(),
+            vk::BufferUsageFlagBits::eUniformTexelBuffer | vk::BufferUsageFlagBits::eStorageTexelBuffer,
+            deviceMemory(),
+            m_memoryPropertyFlags,
+            &m_uniqueBuffer
+        );
+
         m_bufferViews.reserve(m_numPlanes);
         for (uint32_t i = 0; i < m_numPlanes; ++i)
         {
-            auto buffer = Buffer::createFromDeviceMemory(
-                m_device,
-                memorySize(i),
-                vk::BufferUsageFlagBits::eUniformTexelBuffer | vk::BufferUsageFlagBits::eStorageTexelBuffer,
-                deviceMemory(i)
-            );
             m_bufferViews.push_back(BufferView::create(
                 buffer,
                 format(i),
@@ -414,6 +504,7 @@ shared_ptr<BufferView> Image::bufferView(uint32_t plane)
     }
     return m_bufferViews[plane];
 }
+#endif
 
 void Image::importFD(
     const FdDescriptors &descriptors,
@@ -509,25 +600,15 @@ void Image::copyTo(
     auto copyCommands = [&](vk::CommandBuffer commandBuffer) {
         pipelineBarrier(
             commandBuffer,
-            m_imageLayout,
             vk::ImageLayout::eTransferSrcOptimal,
-            m_stage,
             vk::PipelineStageFlagBits::eTransfer,
-            m_accessFlags,
-            vk::AccessFlagBits::eTransferRead,
-            getImageSubresourceRange(),
-            true
+            vk::AccessFlagBits::eTransferRead
         );
         dstImage->pipelineBarrier(
             commandBuffer,
-            dstImage->m_imageLayout,
             vk::ImageLayout::eTransferDstOptimal,
-            dstImage->m_stage,
             vk::PipelineStageFlagBits::eTransfer,
-            dstImage->m_accessFlags,
-            vk::AccessFlagBits::eTransferWrite,
-            dstImage->getImageSubresourceRange(),
-            true
+            vk::AccessFlagBits::eTransferWrite
         );
 
         for (uint32_t i = 0; i < m_numPlanes; ++i)
